#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>

/* the log of masqueraded connections */
#define LOGFILE "/var/log/masquerader"

/* this should be obtained with a getservbyname... */
#define IDENT_PORT 113

/* the timeout on the IDENT query */
#define TIMEOUT 60

/* the logfile */
static FILE *lf = NULL;

/* logfile write locks */
void acquire_lock(void)
{
  struct flock lock;
  lock.l_type = F_WRLCK;
  lock.l_whence = SEEK_SET;
  lock.l_start = 0;
  lock.l_len = 0;
  if (fcntl(fileno(lf),F_SETLKW,&lock) == -1)
    exit(1); /* fatal error (locks not supported) */
}

void release_lock(void)
{
  struct flock lock;
  lock.l_type = F_UNLCK;
  lock.l_whence = SEEK_SET;
  lock.l_start = 0;
  lock.l_len = 0;
  fcntl(fileno(lf),F_SETLK,&lock); /* we ignore errors here... */
}

/* the response of the IDENT query */
static char ident_data[1024];
/* if != "" it contains an error message */
static char error_data[200];
/* argv is used everywhere */
static char **argv_copy;

#define SRC_IP      (argv_copy[1])
#define SRC_PORT    (argv_copy[2])
#define DST_IP      (argv_copy[3])
#define DST_PORT    (argv_copy[4])
#define LOCAL_IP    (argv_copy[5])
#define LOCAL_PORT  (argv_copy[6])
#define OPENED      (argv_copy[7])

void ident_query(void)
{
  static char data[30];
  struct sockaddr_in s;
  int sock, len, count;

  /* use linux transparent proxy support */
  if ((sock = socket(AF_INET,SOCK_STREAM,IPPROTO_TCP)) == -1)
    {
      snprintf(error_data, sizeof(error_data), "socket: %s",strerror(errno));
      return;
    }
  s.sin_family = AF_INET;
  inet_aton(DST_IP, &s.sin_addr); /* spoof the dst address */
  s.sin_port = htons(0);          /* any port is ok */
  if (bind(sock,
	   (struct sockaddr *) &s, 
	   sizeof(struct sockaddr)) == -1)
    {
      snprintf(error_data, sizeof(error_data), "bind: %s",strerror(errno));
      close(sock);
      return;
    }

  /* now connect to the IDENT daemon */
  inet_aton(SRC_IP, &s.sin_addr);
  s.sin_port = htons(IDENT_PORT);
  if (connect(sock, 
	      (struct sockaddr *) &s, 
	      sizeof(struct sockaddr)) == -1)
    {
      snprintf(error_data, sizeof(error_data), "connect: %s",strerror(errno));
      close(sock);
      return;
    }

  /* send query */
  snprintf(data,sizeof(data),"%s , %s\r\n",
	  SRC_PORT,DST_PORT);
  if (write(sock, data, strlen(data)) == -1)
    {
      snprintf(error_data, sizeof(error_data), "write: %s",strerror(errno));
      close(sock);
      return;
    }

  /* wait for response */
  len = 0;
  do 
    {
      count = read(sock, ident_data+len, sizeof(ident_data)-len-2);
      switch (count)
	{
	case -1:
	  close(sock);
	  snprintf(error_data, sizeof(error_data), "read: %s",strerror(errno));
	  ident_data[len++] = '\n';
	  ident_data[len++] = '\0';
	  return;
	case 0:
	  close(sock);
	  snprintf(error_data, sizeof(error_data), 
		   "remote host closed connection");
	  ident_data[len++] = '\n';
	  ident_data[len++] = '\0';
	  return;
	default:
	  len += count;
	  if (len >= sizeof(ident_data)-2)
	    {
	      close(sock);
	      snprintf(error_data, sizeof(error_data), 
		       "IDENT response too long");
	      ident_data[len++] = '\n';
	      ident_data[len++] = '\0';
	      return;
	    }
	}
    } while (ident_data[len-1]!='\n');

  close(sock);
  ident_data[len++]='\0';
}

void write_log(void)
{
  char *c_time;
  time_t t_time;

  alarm(0);

  umask(0077);
  if ((lf = fopen(LOGFILE,"a+")) == NULL) /* create if necessary */
    {
      perror("fopen"); /* this is pretty useless... */
      exit(1); /* fatal error */
    }
  acquire_lock();
  fseek(lf,0,SEEK_END); /* this must be done with the lock */

  t_time = time(NULL);
  c_time = asctime(localtime(&t_time));
  c_time[strlen(c_time)-1] = '\0'; /* chop trailing \n */
  fprintf(lf,"------------ %s -------------\n",
	  c_time);
  fprintf(lf,
	  "from : %s:%s\n"
	  "to   : %s:%s\n"
	  "via  : %s:%s\n"
	  "stat : %s\n",
	  SRC_IP,SRC_PORT,DST_IP,DST_PORT,LOCAL_IP,LOCAL_PORT,OPENED);

  if (ident_data[0] != '\0')
    fprintf(lf,"ident: %s", ident_data);
  if (error_data[0] != '\0')
    fprintf(lf,"error: %s\n", error_data);
  fflush(lf); /* this must be done before releasing the lock */
  release_lock();
  fclose(lf);
}

void sigalrm_handler(int sig)
{
  ident_data[0] = '\0';
  snprintf(error_data,sizeof(error_data),
	   "timeout occoured (%d seconds)", TIMEOUT);
  write_log();
  /* this is not a fatal error: we reported the timeout in the logfile 
   * and that's enough */
  exit(0); 
}

int main(int argc, char **argv)
{
  sigset_t sigset;

  if (argc != 8)
    return 1; /* fatal error */
  
  argv_copy = argv;
  ident_data[0] = error_data[0] = '\0';
  
  sigfillset(&sigset);
  sigdelset(&sigset, SIGKILL);
  sigdelset(&sigset, SIGSTOP);
  sigdelset(&sigset, SIGTERM);
  sigdelset(&sigset, SIGALRM);
  sigprocmask(SIG_SETMASK, &sigset, NULL);
  signal(SIGALRM, sigalrm_handler);
  alarm(TIMEOUT);

  if (OPENED[0] == 'o') /* assume "opened" (new connection) */
    {
      /* do IDENT query */
      ident_query();
    }

  write_log();
  return 0;
}
